#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "vector.h"

vector_t *vector_set(vector_t *vd, double x, double y, double z)
{
  vd->x = x;
  vd->y = y;
  vd->z = z;
  return vd;
}

vector_t *vector_zero(vector_t *vd)
{
  return vector_set(vd, 0.0, 0.0, 0.0);
}

vector_t *vector_rand(vector_t *vd)
{
  return vector_set(vd,
		    ((rand() % 20001) - 10000) * 0.0001,
		    ((rand() % 20001) - 10000) * 0.0001,
		    ((rand() % 20001) - 10000) * 0.0001);
}

vector_t *vector_copy(vector_t *vd, vector_t *v0)
{
  return vector_set(vd, v0->x, v0->y, v0->z);
}

vector_t *vector_add(vector_t *vd, vector_t *v0, vector_t *v1)
{
  return vector_set(vd, v0->x + v1->x, v0->y + v1->y, v0->z + v1->z);
}

vector_t *vector_sub(vector_t *vd, vector_t *v0, vector_t *v1)
{
  return vector_set(vd, v0->x - v1->x, v0->y - v1->y, v0->z - v1->z);
}

vector_t *vector_mul(vector_t *vd, vector_t *v0, double r)
{
  return vector_set(vd, v0->x * r, v0->y * r, v0->z * r);
}

vector_t *vector_div(vector_t *vd, vector_t *v0, double r)
{
  return (r == 0.0) ? NULL : vector_set(vd, v0->x / r, v0->y / r, v0->z / r);
}

vector_t *vector_normalize(vector_t *vd, double r)
{
  double length = vector_length(vd);
  return (length == 0.0) ? NULL : vector_mul(vd, vd, r / length);
}

double vector_square(vector_t *vd)
{
  return vd->x * vd->x + vd->y * vd->y + vd->z * vd->z;
}

double vector_length(vector_t *vd)
{
  return sqrt(vector_square(vd));
}

vector_t *vector_rotate_x(vector_t *vd, vector_t *v0, double angle)
{
  vector_t v;
  double c, s;
  angle *= (2.0 * M_PI) / 360.0;
  c = cos(angle); s = sin(angle);
  v = *v0;
  vd->y = c * v.y - s * v.z;
  vd->z = s * v.y + c * v.z;
  return vd;
}

vector_t *vector_rotate_y(vector_t *vd, vector_t *v0, double angle)
{
  vector_t v;
  double c, s;
  angle *= (2.0 * M_PI) / 360.0;
  c = cos(angle); s = sin(angle);
  v = *v0;
  vd->z = c * v.z - s * v.x;
  vd->x = s * v.z + c * v.x;
  return vd;
}

vector_t *vector_rotate_z(vector_t *vd, vector_t *v0, double angle)
{
  vector_t v;
  double c, s;
  angle *= (2.0 * M_PI) / 360.0;
  c = cos(angle); s = sin(angle);
  v = *v0;
  vd->x = c * v.x - s * v.y;
  vd->y = s * v.x + c * v.y;
  return vd;
}

vector_t *vector_normalize_angle_positive(vector_t *angle)
{
  while (angle->x > 360.0) angle->x -= 360.0;
  while (angle->x <   0.0) angle->x += 360.0;
  while (angle->y > 360.0) angle->y -= 360.0;
  while (angle->y <   0.0) angle->y += 360.0;
  while (angle->z > 360.0) angle->z -= 360.0;
  while (angle->z <   0.0) angle->z += 360.0;
  return angle;
}

vector_t *vector_normalize_angle_sym(vector_t *angle)
{
  vector_normalize_angle_positive(angle);
  if (angle->x > 180.0) angle->x -= 360.0;
  if (angle->y > 180.0) angle->y -= 360.0;
  if (angle->z > 180.0) angle->z -= 360.0;
  return angle;
}

vector_t *vector_calc_angle(vector_t *angle, vector_t *v0, vector_t *updir)
{
  vector_t v, u;

  if ((v0->x == 0.0) && (v0->y == 0.0) && (v0->z == 0.0))
    return vector_zero(angle);

  v = *v0;

  if ((v.y == 0.0) && (v.z == 0.0)) {
    angle->x = 0.0;
  } else {
    angle->x = -(90.0 - atan2(v.z, v.y) * (360.0 / (2.0 * M_PI)));
  }

  vector_rotate_x(&v, &v, -angle->x);

  if ((v.z == 0.0) && (v.x == 0.0)) {
    angle->y = 0.0;
  } else {
    angle->y = atan2(v.x, v.z) * (360.0 / (2.0 * M_PI));
  }

  u = *updir;
  vector_rotate_x(&u, &u, -angle->x);
  vector_rotate_y(&u, &u, -angle->y);

  if ((u.x == 0.0) && (u.y == 0.0)) {
    angle->z = 0.0;
  } else {
    angle->z = atan2(u.y, u.x) * (360.0 / (2.0 * M_PI));
  }

  return vector_normalize_angle_positive(angle);
}
